# TODO: units in plot_pulses

"""
This module contains the PulsedSim class that is used to define a general pulsed experiment with a sequence of pulses and free evolution operations, part of the QuaCAAToo package.
"""

import warnings

import matplotlib.pyplot as plt
import numpy as np
from qutip import Qobj, measurement, mesolve, parallel_map

from ..qsys.qsys import QSys
from .pulse_shapes import square_pulse


class PulsedSim:
    """
    The PulsedSim class is used to define a general pulsed experiment with a sequence of pulses and free evolution operations.
    The class contains methods to add pulses and free evolution operations to the sequence of operations,
    run the experiment, plot the pulse profiles and results of the experiment.
    By default the Hamiltonian is in frequency units and the time is in time units.

    Attributes
    ----------
    system : QSys
        Quantum system object representing the quantum system
    H2 : list(Qobj, function)
        Time dependent sensing Hamiltonian of the system in the form [Qobj, function]
    H0_H2 : list
        List of Hamiltonians for the pulse operation in the form [H0, H2]
    total_time : float
        Total time of the experiment
    variable : np.array
        Variable of the experiment which the results depend on
    variable_name : str
        Name of the variable
    pulse_profiles :list
        List of pulse profiles for plotting purposes, where each element is a list [H1, tarray, pulse_shape, pulse_params]
    results : list
        Results of the experiment to be later generated in the run method
    sequence : callable
        Parallel sequence of operations to be overwritten in predef_seqs and predef_dd_seqs, or defined by the user
    time_steps : int
        Number of time steps for the pulses
    rho : Qobj
        Density matrix of the system
    M : int
        Order of the sequence, if applicable
    pi_pulse_duration : float
        Duration of the pi pulse, if applicable
    free_duration : np.array
        Free evolution times of the sequence, if applicable
    pulse_shape : callable or list(callable)
        Pulse shape function or list of pulse shape functions representing the time modulation of H1
    pulse_params : dict
        Dictionary of parameters for the pulse_shape functions
    options : dict
        Options for the Qutip solver, such as 'nsteps', 'atol', 'rtol', 'order'
    H1 : Qobj or list(Qobj)
        Control Hamiltonian of the system, which can be a single Qobj or a list of Qobjs
    Ht : list
        List of Hamiltonians for the pulse operation in the form [H0, [H1, pulse_shape], H2]
    time_steps : int
        Number of time steps for the pulses, if applicable

    Methods
    -------
    add_free_evolution :
        Adds a free evolution operation to the sequence of operations of the experiment
    _free_evolution :
        Updates the total time of the experiment and applies the time-evolution operator to perform the free evolution operation with the exponential operator
    add_pulse :
        Adds a pulse operation to the sequence of operations of the experiment
    pulse :
        Updates the total time of the experiment, sets the phase for the pulse and calls mesolve from QuTip to perform the pulse operation
    add_delta_pulse :
        Adds a delta pulse with zero duration to the sequence of operations of the experiment
    _delta_pulse :
        Applies the delta pulse operation to the initial state by multiplying the rho attribute with the rotation operator R
    run :
        Runs the pulsed experiment by calling the parallel_map function from QuTip over the variable attribute
    _get_results :
        Gets the results of the experiment from the calculated rho, based on the observable of the system
    measure_qsys :
        Measures the observable over the system, storing the measurement outcome in the results attribute and collapsing rho in the corresponding eigenstate of the observable
    plot_pulses :
        Plots the pulse profiles of the experiment by iterating over the pulse_profiles list and plotting each pulse profile and free evolution
    _check_attr_predef_seqs :
        Checks the common attributes of the PulsedSim object for the predefined sequences and sets them accordingly
    """

    def __init__(self, system, H2=None):
        """
        Initializes a general PulsedSim object with a quantum system, time dependent Hamiltonian and collapse operators.

        Parameters
        ----------
        system : QSys
            Quantum system object representing the quantum system
        H2 : Qobj
            Time dependent sensing Hamiltonian of the system
        """
        if not isinstance(system, QSys):
            raise ValueError("system must be a QSys object")

        self.system = system

        if system.rho0 is not None:
            self.rho = system.rho0.copy()

        # if collapse operators are given, the H0_H2 attributed needs to be set with H0 for the mesolve function
        if self.system.c_ops is not None:
            self.H0_H2 = self.system.H0

        if H2 is None:
            self.H2 = None
        elif H2[0].shape != self.system.H0.shape or not callable(H2[1]):
            raise ValueError(
                "H2 must be a list where the first element is a Qobj of the same shape as H0 and the second element is a time dependent function"
            )
        else:
            self.H2 = H2
            self.H0_H2 = [self.system.H0, self.H2]

        # initialize the rest of the attributes
        self.total_time = 0
        self.variable = None
        self.variable_name = None
        self.pulse_profiles = []
        self.results = []
        self.sequence = None
        self.time_steps = None

    def add_free_evolution(self, duration, options=None):
        """
        Adds a free evolution operation to the sequence of operations of the experiment for a given duration of the free evolution by calling the _free_evolution method.

        Parameters
        ----------
        duration : float or int
            Duration of the free evolution
        options : dict
            Options for the Qutip solver
        """
        # check if duration of the pulse is a positive real number
        if not isinstance(duration, (int, float)) or duration < 0:
            raise ValueError("duration must be a positive real number")

        if options is None:
            options = {}
        elif not isinstance(options, dict):
            raise ValueError("options must be a dictionary of dynamic solver options from Qutip")

        # add the free evolution to the pulse_profiles list
        self.pulse_profiles.append([None, [self.total_time, duration + self.total_time], None, None])

        self._free_evolution(duration, options)

    def _free_evolution(self, duration, options):
        """
        Updates the total time of the experiment and applies the time-evolution operator to the initial state.
        This method should be used internally by other methods, as it does not perform any checks on the input parameters for better performance.
        If the system has collapse operators or time dependent Hamiltonian H2, mesolve is used to perform the free evolution operation.
        Otherwise, the time-evolution operator is applied directlyby the exponential operator.

        Parameters
        ----------
        duration : float or int
            Duration of the free evolution
        """
        if self.system.c_ops is not None or self.H2 is not None:
            self.rho = mesolve(
                self.H0_H2,
                self.rho,
                2 * np.pi * np.linspace(self.total_time, self.total_time + duration, self.time_steps),
                self.system.c_ops,
                e_ops=[],
                options=options,
            ).states[-1]
        elif self.rho.isket:
            self.rho = (-1j * 2 * np.pi * self.system.H0 * duration).expm() * self.rho
        else:
            self.rho = (
                (-1j * 2 * np.pi * self.system.H0 * duration).expm()
                * self.rho
                * ((-1j * 2 * np.pi * self.system.H0 * duration).expm()).dag()
            )

        self.total_time += duration

    def add_pulse(
        self, duration, H1, pulse_shape=square_pulse, pulse_params=None, time_steps=100, options=None
    ):
        """
        Perform variables checks and adds a pulse operation to the sequence of operations of the experiment for a given duration of the pulse,
        control Hamiltonian H1, pulse phase, pulse shape function, pulse parameters and time steps by calling the pulse method.

        Parameters
        ----------
        duration : float or int
            Duration of the pulse
        H1 : Qobj or list(Qobj)
            Control Hamiltonian of the system
        pulse_shape : callable or list(callable)
            Pulse shape function or list of pulse shape functions representing the time modulation of t H1
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions
        time_steps : int
            Number of time steps for the pulses
        options : dict
            Options for the Qutip solver
        """
        # check all the parameters
        if options is None:
            options = {}
        elif not isinstance(options, dict):
            raise ValueError("options must be a dictionary of dynamic solver options from Qutip")

        if not isinstance(time_steps, int) or time_steps <= 0:
            raise ValueError("time_steps must be a positive integer")
        else:
            self.time_steps = time_steps

        if not isinstance(duration, (int, float)) and duration <= 0:
            raise ValueError("duration must be a positive real number")

        if not (
            callable(pulse_shape)
            or (isinstance(pulse_shape, list) and all(callable(p) for p in pulse_shape))
        ):
            raise ValueError("pulse_shape must be a python function or a list of python functions")

        if pulse_params is None:
            pulse_params = {}
        elif not isinstance(pulse_params, dict):
            raise ValueError("pulse_params must be a dictionary of parameters for the pulse function")

        # if the user doesn't provide a phi_t, set it to 0
        if "phi_t" not in pulse_params:
            pulse_params["phi_t"] = 0

        # check if H1 is a Qobj or a list of Qobj with the same dimensions as H0
        if isinstance(H1, Qobj) and H1.shape == self.system.H0.shape:
            # append it to the pulse_profiles list
            self.pulse_profiles.append(
                [
                    H1,
                    np.linspace(self.total_time, self.total_time + duration, self.time_steps),
                    pulse_shape,
                    pulse_params,
                ]
            )
            if self.H2 is None:
                Ht = [self.system.H0, [H1, pulse_shape]]
            else:
                Ht = [self.system.H0, [H1, pulse_shape], self.H2]

        elif (
            isinstance(H1, list)
            and all(isinstance(op, Qobj) and op.shape == self.system.H0.shape for op in H1)
            and len(H1) == len(pulse_shape)
        ):
            self.pulse_profiles = [
                [
                    H1[i],
                    np.linspace(self.total_time, self.total_time + duration, self.time_steps),
                    pulse_shape[i],
                    pulse_params,
                ]
                for i in range(len(H1))
            ]
            if self.H2 is None:
                Ht = [self.system.H0] + [[H1[i], pulse_shape[i]] for i in range(len(H1))]
            else:
                Ht = [self.system.H0] + [[H1[i], pulse_shape[i]] for i in range(len(H1))] + self.H2

        else:
            raise ValueError(
                "H1 must be a Qobj or a list of Qobjs of the same shape as H0 and with the same length as the pulse_shape list"
            )

        # add the pulse operation to the sequence of operations by calling the pulse method
        self._pulse(Ht, duration, options, pulse_params)

    def _pulse(self, Ht, duration, options, core_pulse_params):
        """
        Calls the mesolve function from QuTip to perform the pulse operation with the given Hamiltonian, time array and options,
        by updating the rho attribute of the class with the result of the operation.
        Adds the pulse duration to the total time.
        This method should be used internally by other methods, as it does not perform any checks on the input parameters for better performance.

        Parameters
        ----------
        Ht : list
            List of Hamiltonians for the pulse operation in the form [H0, [H1, pulse_shape]]
        tarray : np.array
            Time array for the pulse operation
        options : dict
            Options for the Qutip solver
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions
        """
        # perform the pulse operation. The time array is multiplied by 2*pi so that [H*t] has units of radians
        self.rho = mesolve(
            Ht,
            self.rho,
            2 * np.pi * np.linspace(self.total_time, self.total_time + duration, self.time_steps),
            self.system.c_ops,
            e_ops=[],
            options=options,
            args=core_pulse_params,
        ).states[-1]

        self.total_time += duration

    def add_delta_pulse(self, R):
        """
        Adds a delta pulse with zero duration to the sequence of operations of the experiment by calling _delta_pulse method.
        For adding a realistic finite length pulse, check add_pulse method.

        Parameters
        ----------
        R : Qobj
            Rotation operator of the delta pulse
        """
        if not isinstance(R, Qobj) or R.shape != self.system.H0.shape:
            raise ValueError("U must be a Qobj of the same shape as H0")

        self.pulse_profiles.append([R, [self.total_time], None, None])
        self._delta_pulse(R)

    def _delta_pulse(self, R):
        """
        Applies the delta pulse operation to the initial state by multiplying the rho attribute with the rotation operator R.
        This method should be used internally by other methods, as it does not perform any checks on the input parameters for better performance.

        Parameters
        ----------
        R : Qobj
            Rotation operator of the delta pulse
        """
        if self.rho.isket:
            self.rho = R * self.rho
        else:
            self.rho = R * self.rho * R.dag()

    def measure_qsys(self, observable=None, tol=None):
        """
        Measures the observable over the system, storing the measurent outcome in the results attribute and collapsing rho in the corresponding eigenstate of the observable.
        If no observable is given, the observable of the qsys is used.
        The rho attribute needs to be normalized before the measurement is performed, in order to avoid numerical errors within the measure_observable method.

        Parameters
        ----------
        observable : Qobj
            Observable to be measured after the sequence of operations
        tol : float
            Tolerance for the measurement, smallest value for the probabilities

        Returns
        -------
        results : float or list
            Measurement outcome of the observable, which can be a float or a list of floats if the observable is a list of Qobjs
        """
        if tol is not None and (not isinstance(tol, (int, float)) or tol < 0):
            raise ValueError("tol must be a positive real number or None")

        self.rho = self.rho.unit()

        if isinstance(observable, Qobj) and observable.shape == self.system.H0.shape:
            if not observable.isherm:
                warnings.warn("Passed observable is not hermitian.")
            self.results, self.rho = measurement.measure_observable(self.rho, observable, tol)

        elif observable is None and (
            isinstance(self.system.observable, Qobj) and self.system.observable.shape == self.system.H0.shape
        ):
            self.results, self.rho = measurement.measure_observable(self.rho, self.system.observable, tol)

        else:
            raise ValueError("observable must be a Qobj of the same shape as rho0, H0 and H1.")

        return self.results.copy()

    def run(self, variable=None, sequence=None, sequence_kwargs=None, map_kw=None):
        """
        Runs the pulsed experiment by calling the parallel_map function from QuTip over the variable attribute.
        The rho attribute is updated.

        Parameters
        ----------
        variable : np.array
            xaxis variable of the plot representing the parameter being changed in the experiment
        sequence : callable
            Sequence of operations to be performed in the experiment
        sequence_kwargs : dict
            Dictionary of arguments to be passed to the sequence function
        map_kw : dict
            Dictionary of options for the parallel_map function from QuTip
        """
        # if no sequence is passed but the PulsedSim has one, uses the attribute sequence
        if sequence is None and self.sequence is not None:
            pass
        # if a sequence is passed, checks if it is a python function and overwrites the attribute
        elif callable(sequence):
            self.sequence = sequence
        else:
            raise ValueError("sequence must be a python function with a list operations returning a number")

        # check if a variable was passed by the user, if it is numpy array overwrite the variable attribute
        if isinstance(variable, np.ndarray):
            self.variable = variable
        elif variable is None and len(self.variable) != 0:
            pass
        else:
            raise ValueError("variable must be a numpy array")

        # check if map_kw and sequence_kwargs are None or dictionaries
        if map_kw is None:
            map_kw = {"num_cpus": None}
        elif not isinstance(map_kw, dict):
            raise ValueError(
                "map_kw must be a dictionary of options for the parallel_map function from QuTip"
            )

        if sequence_kwargs is None:
            sequence_kwargs = {}
        elif not isinstance(sequence_kwargs, dict):
            raise ValueError(
                "sequence_args must be a dictionary of arguments to be passed to the sequence function"
            )

        # the rho attribute needs to be reset to the initial state, so it doesnt run over the previous simulation
        self.rho = self.system.rho0.copy()

        # run the experiment by calling the parallel_map function from QuTip over the variable attribute
        self.rho = parallel_map(self.sequence, self.variable, task_kwargs=sequence_kwargs, map_kw=map_kw)

        self._get_results()

    def _get_results(self):
        """
        Gets the results of the experiment from the calculated rho, based on the observable of the system.
        The results are stored in the results attribute of the class.
        """
        if self.rho[0].isket:
            if isinstance(self.system.observable, Qobj):
                # np.real is used to ensure no imaginary components will be attributed to results
                self.results = np.array(
                    [np.real(rho.dag() * self.system.observable * rho) for rho in self.rho]
                )
            elif isinstance(self.system.observable, list):
                self.results = [
                    np.array([np.real(rho.dag() * observable * rho) for rho in self.rho])
                    for observable in self.system.observable
                ]
        elif isinstance(self.system.observable, Qobj):
            self.results = np.array([np.real((rho * self.system.observable).tr()) for rho in self.rho])
        elif isinstance(self.system.observable, list):
            self.results = [
                np.array([np.real((rho * observable).tr()) for rho in self.rho])
                for observable in self.system.observable
            ]

    def plot_pulses(self, figsize=(6, 4), xlabel=None, ylabel="Pulse Intensity", title="Pulse Profiles"):
        """
        Plots the pulse profiles of the experiment by iterating over the pulse_profiles list and plotting each pulse profile and free evolution.

        Parameters
        ----------
        figsize : tuple
            Size of the figure to be passed to matplotlib.pyplot
        xlabel : str
            Label of the x-axis
        ylabel : str
            Label of the y-axis
        title : str
            Title of the plot
        """
        if not (isinstance(figsize, tuple) or len(figsize) == 2):
            raise ValueError("figsize must be a tuple of two positive floats")

        if xlabel is None:
            xlabel = self.variable_name
        elif not isinstance(xlabel, str):
            raise ValueError("xlabel must be a string")

        fig, ax = plt.subplots(1, 1, figsize=figsize)

        # iterate over all operations in the sequence
        for itr_pulses in range(len(self.pulse_profiles)):
            # if the pulse profile is a free evolution, plot a horizontal line at 0
            if self.pulse_profiles[itr_pulses][0] is None:
                ax.plot(
                    self.pulse_profiles[itr_pulses][1],
                    [0, 0],
                    label="Free Evolution",
                    lw=2,
                    alpha=0.7,
                    color="C0",
                )

            # if the pulse has operators, plot the pulse profiles
            elif isinstance(self.pulse_profiles[itr_pulses][0], Qobj):
                ax.plot(
                    self.pulse_profiles[itr_pulses][1],
                    self.pulse_profiles[itr_pulses][2](
                        2 * np.pi * self.pulse_profiles[itr_pulses][1], **self.pulse_profiles[itr_pulses][3]
                    ),
                    label="H1",
                    lw=2,
                    alpha=0.7,
                    color="C1",
                )

            elif isinstance(self.pulse_profiles[itr_pulses][0], list):
                for itr_op in range(len(self.pulse_profiles[itr_pulses])):
                    ax.plot(
                        self.pulse_profiles[itr_pulses][itr_op][1],
                        self.pulse_profiles[itr_pulses][itr_op][2](
                            2 * np.pi * self.pulse_profiles[itr_pulses][itr_op][1],
                            **self.pulse_profiles[itr_pulses][itr_op][3],
                        ),
                        label=f"H1_{itr_op}",
                        lw=2,
                        alpha=0.7,
                        color=f"C{2 + itr_op}",
                    )

        ax.set_xlim(0, self.total_time)
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_title(title)

        # make sure that the legend only shows unique labels.
        # Adapted from user Julien J in https://stackoverflow.com/questions/19385639/duplicate-items-in-legend-in-matplotlib/40870637#40870637
        handles, labels = ax.get_legend_handles_labels()
        unique_legend = [(h, l) for i, (h, l) in enumerate(zip(handles, labels)) if l not in labels[:i]]
        ax.legend(*zip(*unique_legend), loc="upper right", bbox_to_anchor=(1.2, 1))

    def _check_attr_predef_seqs(
        self,
        H1,
        pulse_shape,
        pulse_params,
        options,
        time_steps,
        free_duration,
        pi_pulse_duration,
        M,
        projection_pulse,
    ):
        """
        Checks the commom attributes of the PulsedSim object for the predefined sequences and sets them accordingly.

        Parameters
        ----------
        H1 : Qobj or list(Qobj)
            Control Hamiltonian of the system
        pulse_shape : callable or list(callable)
            Pulse shape function or list of pulse shape functions representing the time modulation of H1
        pulse_params : dict
            Dictionary of parameters for the pulse_shape functions
        options : dict
            Options for the Qutip solver
        time_steps : int
            Number of time steps for the pulses, if applicable
        free_duration : np.array
            Free evolution times of the sequence, if applicable
        pi_pulse_duration : float
            Duration of the pi pulse, if applicable
        M : int
            Order of the sequence, if applicable
        projection_pulse : bool
            Whether the sequence contains a final projection pulse or not, if applicable
        """
        # check whether pulse_shape is a python function or a list of python functions and if it is, assign it to the object
        if callable(pulse_shape) or (
            isinstance(pulse_shape, list) and all(callable(pulse_shape) for pulse_shape in pulse_shape)
        ):
            self.pulse_shape = pulse_shape
        else:
            raise ValueError("pulse_shape must be a python function or a list of python functions")

        # check whether pulse_params is a dictionary and if it is, assign it to the object
        if pulse_params is None:
            self.pulse_params = {}
        elif isinstance(pulse_params, dict):
            self.pulse_params = pulse_params
        else:
            raise ValueError(
                "pulse_params must be a dictionary or a list of dictionaries of parameters for the pulse function"
            )

        # if phi_t is not in the pulse_params dictionary, assign it as 0
        if "phi_t" not in self.pulse_params:
            self.pulse_params["phi_t"] = 0

        # check whether options is a dictionary of solver options from Qutip and if it is, assign it to the object
        if options is None:
            self.options = {}
        elif isinstance(options, dict):
            self.options = options
        else:
            raise ValueError("options must be a dictionary of dynamic solver options from Qutip")

        # check whether H1 is a Qobj or a list of Qobjs of the same shape as H0 and with the same length as the pulse_shape list and if it is, assign it to the object
        if isinstance(H1, Qobj) and H1.shape == self.system.H0.shape:
            self.H1 = H1
            if self.H2 is None:
                self.Ht = [self.system.H0, [H1, pulse_shape]]
            else:
                self.Ht = [self.system.H0, [H1, pulse_shape], self.H2]
                self.H0_H2 = [self.system.H0, self.H2]

        elif (
            isinstance(H1, list)
            and all(isinstance(op, Qobj) and op.shape == self.system.H0.shape for op in H1)
            and len(H1) == len(pulse_shape)
        ):
            self.H1 = H1
            if self.H2 is None:
                self.Ht = [self.system.H0] + [[H1[i], pulse_shape[i]] for i in range(len(H1))]
            else:
                self.Ht = [self.system.H0] + [[H1[i], pulse_shape[i]] for i in range(len(H1))] + self.H2
                self.H0_H2 = [self.system.H0, self.H2]

        else:
            raise ValueError(
                "H1 must be a Qobj or a list of Qobjs of the same shape as H0 with the same length as the pulse_shape list"
            )

        # check whether time_steps is a positive integer and if it is, assign it to the object
        if not isinstance(time_steps, int) or time_steps <= 0:
            raise ValueError("time_steps must be a positive integer")
        else:
            self.time_steps = time_steps

        # check whether free_duration is a numpy array of real and positive elements and if it is, assign it to the object
        if free_duration is None:
            pass
        elif (
            not isinstance(free_duration, (np.ndarray, list))
            or not np.all(np.isreal(free_duration))
            or not np.all(np.greater_equal(free_duration, 0))
        ):
            raise ValueError("free_duration must be a numpy array with real positive elements")
        else:
            self.variable = free_duration
            self.variable_name = f"Tau (1/{self.system.units_H0})"

        # check whether pi_pulse_duration is a positive real number and if it is, assign it to the object
        if pi_pulse_duration is None:
            pass
        elif (
            not isinstance(pi_pulse_duration, (int, float))
            or pi_pulse_duration <= 0
            or pi_pulse_duration > free_duration[0]
        ):
            warnings.warn(
                "pulse_duration must be a positive real number and pi_pulse_duration must be smaller than the free evolution time, otherwise pulses will overlap"
            )
            self.pi_pulse_duration = pi_pulse_duration
        else:
            self.pi_pulse_duration = pi_pulse_duration

        # check whether M is a positive integer and if it is, assign it to the object
        if M is None:
            pass
        elif not isinstance(M, int) or M <= 0:
            raise ValueError("M must be a positive integer")
        else:
            self.M = M

        # check whether projection_pulse is a boolean and if it is, assign it to the object
        if projection_pulse is None:
            pass
        elif isinstance(projection_pulse, bool):
            self.projection_pulse = projection_pulse
        else:
            raise ValueError("projection_pulse must be a boolean")



# TODO: think on a better strategy for the plot_energy_B0 function
# TODO: implement eV units conversion to frequencies

"""
This module contains the plot_energy_B0 function, compose_sys function and the QSys class.
"""

import warnings

import matplotlib.pyplot as plt
import numpy as np
from qutip import Qobj, basis, qeye, tensor


def compose_sys(qsys1, qsys2):
    """
    Takes two quantum systems and returns the composed system by performing tensor products of the two,
    after checking if all parameters are valid.

    Parameters
    ----------
    qsys1 : QSys
        First quantum system.
    qsys2 : QSys
        Second quantum system.

    Returns
    -------
    qsys1 X qsys2 : QSys
        Composed quantum system.
    """
    if not isinstance(qsys1, QSys) or not isinstance(qsys2, QSys):
        raise ValueError("Both qsys1 and qsys2 must be instances of the QSys class.")

    if qsys1.units_H0 != qsys2.units_H0:
        warnings.warn("The two systems have different units.")

    if isinstance(qsys1.H0, Qobj) and isinstance(qsys2.H0, Qobj):
        H0 = tensor(qsys1.H0, qeye(qsys2.H0.dims[0])) + tensor(qeye(qsys1.H0.dims[0]), qsys2.H0)
    else:
        raise ValueError("Both Hamiltonians must be Qobj.")

    if qsys1.rho0.isherm and qsys2.rho0.isherm or qsys1.rho0.isket and qsys2.rho0.isket:
        rho0 = tensor(qsys1.rho0, qsys2.rho0).unit()
    else:
        rho0 = None

    if qsys1.observable is not None and qsys2.observable is not None:
        if isinstance(qsys1.observable, Qobj) and isinstance(qsys2.observable, Qobj):
            observable = tensor(qsys1.observable, qsys2.observable)
        elif isinstance(qsys1.observable, list) and isinstance(qsys2.observable, list):
            observable = [tensor(obs1, qeye(qsys2.H0.shape[0])) for obs1 in qsys1.observable] + [
                tensor(qeye(qsys1.H0.shape[0]), obs2) for obs2 in qsys2.observable
            ]
        else:
            raise ValueError("Both observables must be Qobj or None")
    else:
        observable = None

    if qsys1.c_ops is None and qsys2.c_ops is None:
        c_ops = None
    elif qsys1.c_ops is not None and qsys2.c_ops is None:
        if isinstance(qsys1.c_ops, Qobj):
            c_ops = tensor(qsys1.c_ops, qeye(qsys2.H0.shape[0]))
        elif isinstance(qsys1.c_ops, list):
            c_ops = [tensor(op1, qeye(qsys2.H0.shape[0])) for op1 in qsys1.c_ops]
    elif qsys1.c_ops is None and qsys2.c_ops is not None:
        if isinstance(qsys2.c_ops, Qobj):
            c_ops = tensor(qeye(qsys1.H0.shape[0]), qsys2.c_ops)
        elif isinstance(qsys2.c_ops, list):
            c_ops = [tensor(qeye(qsys1.H0.shape[0]), op2) for op2 in qsys2.c_ops]
    elif qsys1.c_ops is not None and qsys2.c_ops is not None:
        if isinstance(qsys1.c_ops, Qobj) and isinstance(qsys2.c_ops, Qobj):
            c_ops = [
                tensor(qsys1.c_ops, qeye(qsys2.H0.shape[0])),
                tensor(qeye(qsys1.H0.shape[0]), qsys2.c_ops),
            ]
        elif isinstance(qsys1.c_ops, list) and isinstance(qsys2.c_ops, list):
            c_ops = [tensor(op1, qeye(qsys2.H0.shape[0])) for op1 in qsys1.c_ops] + [
                tensor(qeye(qsys1.H0.shape[0]), op2) for op2 in qsys2.c_ops
            ]
        elif isinstance(qsys1.c_ops, Qobj) and isinstance(qsys2.c_ops, list):
            c_ops = [tensor(qsys1.c_ops, qeye(qsys2.H0.shape[0]))] + [
                tensor(qeye(qsys1.H0.shape[0]), op2) for op2 in qsys2.c_ops
            ]
        elif isinstance(qsys1.c_ops, list) and isinstance(qsys2.c_ops, Qobj):
            c_ops = [tensor(op1, qeye(qsys2.H0.shape[0])) for op1 in qsys1.c_ops] + [
                tensor(qeye(qsys1.H0.shape[0]), qsys2.c_ops)
            ]
    else:
        raise ValueError("Both collapse operators must be Qobj, a list of Qobj or None")

    return QSys(H0, rho0, c_ops, observable, qsys1.units_H0)


def plot_energy_B0(B0, H0, figsize=(6, 4), energy_lim=None, xlabel="Magnetic Field", ylabel="Energy (MHz)"):
    """
    Plots the energy levels of a Hamiltonian as a function of a magnetic field B0.

    Parameters
    ----------
    B0 : list or numpy.ndarray
        List of magnetic fields.
    H0 : list of Qobj
        List of Hamiltonians.
    figsize : tuple
        Size of the figure.
    energy_lim : tuple
        Limits of the energy levels.
    xlabel : str
        Label of the x-axis.
    ylabel : str
        Label of the y-axis.
    """
    if not (isinstance(figsize, tuple) or len(figsize) == 2):
        raise ValueError("figsize must be a tuple of two positive floats")

    if not isinstance(B0, (np.ndarray, list)) and all(isinstance(b, (int, float)) for b in B0):
        raise ValueError("B0 must be a list or a numpy array of real numbers")

    if not isinstance(H0, list) or not all(isinstance(h, Qobj) for h in H0) or len(H0) != len(B0):
        raise ValueError("H0 must be a list of Qobj of the same size as B0")

    energy_levels = []

    H0_0 = H0[0].eigenenergies()[0]

    # iterate over all the Hamiltonians and calculate the energy levels
    for itr_B0 in range(len(B0)):
        H0_eig = H0[itr_B0].eigenenergies()
        energy_levels.append(H0_eig - H0_0)

    fig, ax = plt.subplots(1, 1, figsize=figsize)

    for itr_e in range(len(energy_levels[0])):
        ax.plot(B0, [energy_levels[itr_B0][itr_e] for itr_B0 in range(len(B0))])

    if isinstance(energy_lim, tuple) and len(energy_lim) == 2:
        ax.set_ylim(energy_lim)

    if isinstance(xlabel, str):
        ax.set_xlabel(xlabel)
    if isinstance(ylabel, str):
        ax.set_ylabel(ylabel)

    fig.suptitle("Energy Levels")


####################################################################################################


class QSys:
    """
    The QSys class defines a general quantum system and contains a method for plotting the energy levels of the Hamiltonian.

    Attributes
    ----------
    H0 : Qobj or numpy.ndarray
        Time-independent internal Hamiltonian of the system.
    rho0 : Qobj, numpy.ndarray, or int
        Initial state of the system.
    c_ops : Qobj or list of Qobj
        List of collapse operators.
    observable : Qobj or list of Qobj
        Observable to be measured.
    units_H0 : str
        Units of the Hamiltonian.
    energy_levels : numpy.ndarray
        Energy levels of the Hamiltonian.
    eigenstates : numpy.ndarray of Qobj
        Eigenstates of the Hamiltonian.
    dim_add_spin : int
        Dimension of the added spin if the add_spin method is used.

    Methods
    -------
    _get_energy_levels
        Calculates the eigenenergies of the Hamiltonian and sets the energy_levels and eigenstates attributes.
    plot_energy
        Plots the energy levels of the Hamiltonian.
    add_spin
        Adds another spin to the system's Hamiltonian and updates the system accordingly.
    truncate
        Truncates the quantum system by removing the states specified by the indexes from the system
    """

    def __init__(self, H0, rho0=None, c_ops=None, observable=None, units_H0=None):
        """
        Construct the QSys class. It initializes the system with the Hamiltonian, the initial state, the collapse operators and the observable.
        Checking all inputs and calculating the energy levels of the Hamiltonian.

        Parameters
        ----------
        H0 : Qobj/array
            Time-independent internal Hamiltonian of the system
        rho0 : Qobj/array/int
            Initial state of the system. Can be a Qobj, an array or an index number indicating the system eigenstates
        c_ops : list(Qobj)
            List of collapse operators
        observable : Qobj or list(Qobj)
            Observable to be measured
        units_H0 : str
            Units of the Hamiltonian
        """
        # if the units are in frequency, assign the Hamiltonian as it is
        if units_H0 is None:
            self.units_H0 = "MHz"
            warnings.warn("No units supplied, assuming default value of MHz.")
        elif units_H0 in ["MHz", "GHz", "kHz", "eV"]:
            self.units_H0 = units_H0
        else:
            raise ValueError(
                f"Invalid value for units_H0. Expected either units of frequencies or 'eV', got {units_H0}. The Hamiltonian will be considered in MHz."
            )

        self.H0 = Qobj(H0)

        if not self.H0.isherm:
            warnings.warn("Passed H0 is not a hermitian object.")

        self._get_energy_levels()

        # check if rho0 is correctly defined
        if rho0 is None:
            warnings.warn("Initial state not provided.")
        elif rho0 in range(len(self.eigenstates)):
            self.rho0 = self.eigenstates[rho0]  # In this case the initial state is the i-th energy state
        elif (
            Qobj(rho0).isket
            and Qobj(rho0).shape[0] == H0.shape[0]
            or Qobj(rho0).isherm
            and Qobj(rho0).shape == H0.shape
        ):
            self.rho0 = Qobj(rho0)
        else:
            raise ValueError(
                "rho0 must be a Qobj, None or an index number indicating the system eigenstates"
            )

        # check if observable is not None, or if it is a Qobj of the same dimension as H0 or a list of Qobj
        if observable is None:
            self.observable = None
        elif isinstance(observable, (Qobj, np.ndarray)) and observable.shape == H0.shape:
            self.observable = observable
            if not observable.isherm:
                warnings.warn("Passed observable is not hermitian.")
        elif (
            isinstance(observable, list)
            and all(isinstance(obs, (Qobj, np.ndarray)) for obs in observable)
            and all(obs.shape == H0.shape for obs in observable)
        ):
            self.observable = observable
            if not all(obs.isherm for obs in observable):
                warnings.warn("Passed observables are not hermitian.")
        else:
            raise ValueError(
                "Invalid value for observable. Expected a Qobj or a list of Qobj of the same dimensions as H0."
            )

        # check if c_ops is a list of Qobj with the same dimensions as H0
        if c_ops is None or isinstance(c_ops, Qobj) and c_ops.shape == self.H0.shape:
            self.c_ops = c_ops
        elif isinstance(c_ops, list):
            if all(isinstance(op, (Qobj, np.ndarray)) and op.shape == self.H0.shape for op in c_ops):
                self.c_ops = c_ops
            else:
                raise ValueError("All items in c_ops must be Qobj with the same dimensions as H0")
        else:
            raise ValueError("c_ops must be a list of Qobj or None")

    def _get_energy_levels(self):
        """
        Calculates the eigenenergies of the Hamiltonian and subtract the ground state energy from all of them to get the lowest level at 0.
        Sets the energy_levels and eigenstates attributes of the class.
        """
        H_eig = self.H0.eigenenergies()
        self.energy_levels = H_eig - H_eig[0]

        self.eigenstates = self.H0.eigenstates()[1]

    def plot_energy(self, figsize=(2, 6), energy_lim=None):
        """
        Plots the energy levels of the Hamiltonian defined in the system.

        Parameters
        ----------
        figsize : tuple
            Size of the figure
        energy_lim : list
            Limits of the energy levels
        """
        if not (isinstance(figsize, tuple) or len(figsize) == 2):
            raise ValueError("figsize must be a tuple of two positive floats")

        fig, ax = plt.subplots(1, 1, figsize=figsize)

        for itr in range(self.energy_levels.size):
            ax.axhline(y=self.energy_levels[itr], lw=2)

        if self.units_H0 is not None:
            ax.set_ylabel(f"Energy ({self.units_H0})")
        else:
            ax.set_ylabel("Energy")

        ax.get_xaxis().set_visible(False)
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

        if energy_lim is None:
            ax.set_ylim(-0.05 * self.energy_levels[-1], 1.05 * self.energy_levels[-1])
        elif len(energy_lim) == 2:
            ax.set_ylim(energy_lim[0], energy_lim[1])
        else:
            raise ValueError("freq_lim must be a tuple of two floats")

    def add_spin(self, H_spin):
        """
        Adds another spin to the system's Hamiltonian, given a Hamiltonian for the new spin.
        Updates the time-independent Hamiltonian, the energy levels, the initial state, the observable and the collapse operators accordingly.

        Parameters
        ----------
        H_spin : Qobj
            Hamiltonian of the new spin
        """
        if not isinstance(H_spin, Qobj):
            raise ValueError(
                "H_spin must be a Qobj in the form of a tensor with the original Hamiltonian H0."
            )

        if not Qobj(H_spin).isherm:
            warnings.warn("Passed H_spin is not a hermitian object.")

        self.dim_add_spin = int(H_spin.shape[0] / self.H0.shape[0])

        if self.dim_add_spin <= 1:
            raise ValueError("H_spin must be a Qobj with a dimension higher than H0.")

        self.H0 = tensor(self.H0, qeye(self.dim_add_spin)) + H_spin
        self._get_energy_levels()

        if self.rho0.isherm:
            self.rho0 = tensor(self.rho0, qeye(self.dim_add_spin)).unit()
        elif self.rho0.isket:
            self.rho0 = tensor(self.rho0, basis(self.dim_add_spin, 0)).unit()

        if self.observable is not None:
            if isinstance(self.observable, Qobj):
                self.observable = tensor(self.observable, qeye(self.dim_add_spin))
            elif isinstance(self.observable, list) and all(isinstance(obs, Qobj) for obs in self.observable):
                self.observable = [tensor(obs, qeye(self.dim_add_spin)) for obs in self.observable]

        if self.c_ops is not None:
            if isinstance(self.c_ops, Qobj):
                self.c_ops = tensor(self.c_ops, qeye(self.dim_add_spin))
            elif isinstance(self.c_ops, list) and all(isinstance(op, Qobj) for op in self.c_ops):
                self.c_ops = [tensor(op, qeye(self.dim_add_spin)) for op in self.c_ops]

    def truncate(self, indexes):
        """
        Truncantes the quantum system by removing the states specified by the indexes from the Hamiltonian,
        the initial state, the observable and the collapse operators. For example, if S=1 and the user wants to remove the ms=-1 state,
        the indexes is set to 0 and the qsys is truncated to the ms=0 and ms=+1 subspace. The method uses the numpy.delete function
        to remove the specified indexes from the Hamiltonian and the initial state.

        Parameters
        ----------
        indexes : int or list of int
            Index or list of indexes to be removed from the system.
        """
        if isinstance(indexes, int):
            if indexes < 0 or indexes >= self.H0.shape[0]:
                raise ValueError("sel must be a valid index of the Hamiltonian.")
        if isinstance(indexes, (list, np.array)):
            if not all(isinstance(i, int) for i in indexes) and not all(
                0 <= i < self.H0.shape[0] for i in indexes
            ):
                raise ValueError("All elements in sel must be valid indices of the Hamiltonian.")
        else:
            raise ValueError("sel must be an integer or a list of integers.")

        self.H0 = Qobj(np.delete(np.delete(self.H0.full(), indexes, axis=0), indexes, axis=1))
        self._get_energy_levels()

        if self.observable is not None:
            if isinstance(self.observable, Qobj):
                self.observable = Qobj(
                    np.delete(np.delete(self.observable.full(), indexes, axis=0), indexes, axis=1)
                )
            elif isinstance(self.observable, list):
                self.observable = [
                    Qobj(np.delete(np.delete(obs.full(), indexes, axis=0), indexes, axis=1))
                    for obs in self.observable
                ]

        if self.rho0 is not None:
            if self.rho0.isket:
                self.rho0 = Qobj(np.delete(self.rho0.full(), indexes, axis=0)).unit()
            else:
                self.rho0 = Qobj(
                    np.delete(np.delete(self.rho0.full(), indexes, axis=0), indexes, axis=1)
                ).unit()

        if self.c_ops is not None:
            if isinstance(self.c_ops, Qobj):
                self.c_ops = Qobj(np.delete(np.delete(self.c_ops.full(), indexes, axis=0), indexes, axis=1))
            elif isinstance(self.c_ops, list):
                self.c_ops = [
                    Qobj(np.delete(np.delete(op.full(), indexes, axis=0), indexes, axis=1))
                    for op in self.c_ops
                ]
